// Software developed for the Spring 2016 Brown University course
// ENGN 2502 3D Photography
// Copyright (c) 2016, Daniel Moreno and Gabriel Taubin
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the Brown University nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL DANIEL MORENO NOR GABRIEL TAUBIN BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "CamCalibPanel.hpp"

#include <QShowEvent>
#include <QHideEvent>
#include <QMessageBox>
#include <QSettings>
#include <QPainter>
#include <QTimer>
#include <QDir>

#include "Application.hpp"

#include "gui_util.hpp"

#ifdef HAVE_HOMEWORK
# include <homework/homework2.hpp>
#endif //HAVE_HOMEWORK

bool CamCalibPanel::_registered = CamCalibPanel::registerPanel();
bool CamCalibPanel::registerPanel()
{
   qDebug("CamCalibPanel: register");
   return MainWindow::registerPanel("Camera Calibration", [](QWidget *parent) -> QWidget*{ return new CamCalibPanel(parent); });
   return true;
}

CamCalibPanel::CamCalibPanel(QWidget * parent, Qt::WindowFlags f) :
  QWidget(parent, f),
  _calib(),
  _model(),
  _reprojection()
{
  setupUi(this);

  QSettings config;

  //load other settings
  loadSettings();

  //list view model
  imageList->setModel(&_model);
  auto selectionModel = imageList->selectionModel();
  connect(selectionModel, &QItemSelectionModel::currentChanged, this, &CamCalibPanel::_on_currentImageChanged);

  //progress bar
  reprojectProgress->setVisible(false);

  qDebug(" -- CamCalibPanel created -- ");
}

CamCalibPanel::~CamCalibPanel()
{
  saveSettings();
  qDebug(" -- CamCalibPanel destroyed -- ");
}

void CamCalibPanel::loadSettings(void)
{
  QSettings config;

  displayCornersCheck->setChecked(config.value("CamCalibPanel/displayCorners", false).toBool());
  displayReprojectionCheck->setChecked(config.value("CamCalibPanel/displayReprojection", false).toBool());
}

void CamCalibPanel::saveSettings(void)
{
  QSettings config;

  config.setValue("CamCalibPanel/displayCorners", displayCornersCheck->isChecked());
  config.setValue("CamCalibPanel/displayReprojection", displayReprojectionCheck->isChecked());
}

void CamCalibPanel::showEvent(QShowEvent * e)
{
  if (e && !e->spontaneous())
  {
    auto mainWin = getApp()->getMainWindow();
    auto glWidget = mainWin->getGLWidget();
    glWidget->clearImage();
    glWidget->setRotation(0);
    connect(mainWin, &MainWindow::workDirectoryChanged, this, &CamCalibPanel::_on_workDirectoyChanged);
    loadMatlabCalib();
  }
}

void CamCalibPanel::hideEvent(QHideEvent * e)
{
  if (e && !e->spontaneous())
  {
    auto mainWin = getApp()->getMainWindow();
    disconnect(mainWin, &MainWindow::workDirectoryChanged, this, &CamCalibPanel::_on_workDirectoyChanged);
  }
}

void CamCalibPanel::_on_workDirectoyChanged()
{
  loadMatlabCalib();
}

void CamCalibPanel::updateImageFiles()
{
  //get image files
  QStringList fileList;
  auto const& imageData = _calib.imageData;
  foreach(auto & data, imageData)
  {
    fileList.append(data.imageName);
  }

  //update the model
  gui_util::updateFilesItems(&_model, fileList);
}

void CamCalibPanel::_on_currentImageChanged(const QModelIndex & current, const QModelIndex & previous)
{
  Q_UNUSED(current);
  Q_UNUSED(previous);
  updateImage(gui_util::loadCurrentImage(imageList));
}

void CamCalibPanel::on_displayReprojectionCheck_stateChanged()
{
  updateImage(gui_util::loadCurrentImage(imageList, false));
}

void CamCalibPanel::on_displayCornersCheck_stateChanged()
{
  updateImage(gui_util::loadCurrentImage(imageList, false));
}

void CamCalibPanel::on_reprojectCurrentButton_clicked()
{
  //get selected item
  QModelIndex modelIndex;
  auto item = gui_util::getCurrentItem(imageList, &modelIndex);
  if (!item)
  { //no item selected
    QMessageBox::critical(this, "Error", "No item is selected.");
    return;
  }

  auto mainWin = getApp()->getMainWindow();

  //disable gui
  mainWin->setChangesEnabled(false);
  imageList->setEnabled(false);
  displayCornersCheck->setEnabled(false);
  displayReprojectionCheck->setEnabled(false);
  reprojectCurrentButton->setEnabled(false);
  reprojectAllButton->setEnabled(false);
  calibGroup->setEnabled(false);

  //busy cursor
  QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
  QApplication::processEvents();

  //do reprojection
  reprojectImage(modelIndex.row());

  //update gui
  updateImage(gui_util::loadCurrentImage(imageList, false));
  updateMinMaxError();

  //enable gui
  mainWin->setChangesEnabled(true);
  imageList->setEnabled(true);
  displayCornersCheck->setEnabled(true);
  displayReprojectionCheck->setEnabled(true);
  reprojectCurrentButton->setEnabled(true);
  reprojectAllButton->setEnabled(true);
  calibGroup->setEnabled(true);

  //restore cursor
  QApplication::restoreOverrideCursor();
}

void CamCalibPanel::on_reprojectAllButton_clicked()
{
  auto mainWin = getApp()->getMainWindow();

  //disable gui
  mainWin->setChangesEnabled(false);
  imageList->setEnabled(false);
  displayReprojectionCheck->setEnabled(false);
  reprojectCurrentButton->setEnabled(false);
  reprojectAllButton->setEnabled(false);
  calibGroup->setEnabled(false);

  //busy cursor
  QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
  QApplication::processEvents();

  int rows = _model.rowCount();
  reprojectProgress->setMaximum(rows);
  reprojectProgress->setVisible(true);
  for (int i = 0; i < rows; ++i)
  {
    auto item = _model.item(i, 0);
    if (!item)
    { //invalid item
      QMessageBox::critical(this, "Error", "Invalid item.");
      break;
    }

    //do reprojection
    reprojectImage(i);

    //update progress bar
    reprojectProgress->setValue(i);
    QApplication::processEvents();
  }
  reprojectProgress->setVisible(false);

  //update gui
  updateImage(gui_util::loadCurrentImage(imageList, false));
  updateMinMaxError();

  //enable gui
  mainWin->setChangesEnabled(true);
  imageList->setEnabled(true);
  displayReprojectionCheck->setEnabled(true);
  reprojectCurrentButton->setEnabled(true);
  reprojectAllButton->setEnabled(true);
  calibGroup->setEnabled(true);

  //restore cursor
  QApplication::restoreOverrideCursor();
}

void CamCalibPanel::reprojectImage(int imageIndex)
{
  if (imageIndex<0 || imageIndex>=_calib.imageData.size())
  {
    return;
  }

  auto const& K = _calib.K;
  auto const& kc = _calib.kc;
  auto const& imageData = _calib.imageData[imageIndex];
  auto R = rodrigues(imageData.rotVec);
  auto const& T = imageData.T;
  auto const& worldPoints = imageData.worldPoints;
  auto const& imagePoints = imageData.imagePoints;

  int imageCount = static_cast<int>(_calib.imageCount);
  if (_reprojection.size()!=imageCount)
  {
    _reprojection.resize(imageCount);
  }

  auto & reproj = _reprojection[imageIndex];
  reproj.imagePoints.clear();
  reproj.rmsError = hw2::reprojectPoints(K, kc, R, T, worldPoints, imagePoints, reproj.imagePoints);
}

void CamCalibPanel::updateImage(QImage image)
{
  auto mainWin = getApp()->getMainWindow();
  if (!mainWin)
  {
    return;
  }
  auto glWidget = mainWin->getGLWidget();

  if (image.isNull())
  {
    glWidget->clearImage();
    return;
  }

  QPainter painter(&image);
  int index = image.text("index").toInt();
  int radius = 8, diameter = 2*radius;

  //draw corners
  if (displayCornersCheck->isChecked() && index<_calib.imageData.size())
  {
    painter.setPen(Qt::yellow);
    auto const& imagePoints = _calib.imageData[index].imagePoints;
    foreach(auto const& p, imagePoints)
    {
      painter.drawEllipse(p.x()-radius, p.y()-radius, diameter, diameter);
      painter.drawPoint(p.x(), p.y());
    }
  }

  //draw reprojected points
  if (displayReprojectionCheck->isChecked() && index<_reprojection.size())
  {
    painter.setPen(Qt::red);
    auto const& imagePoints = _reprojection[index].imagePoints;
    foreach(auto const& p, imagePoints)
    {
      painter.drawLine(p.x()-radius, p.y(), p.x()+radius, p.y());
      painter.drawLine(p.x(), p.y()-radius, p.x(), p.y()+radius);
    }
  }

  //update rms error lineEdit
  double rmsError = 0.0;
  if (index<_reprojection.size())
  {
    rmsError = _reprojection[index].rmsError;
  }
  rmsErrorLine->setText(QString::number(rmsError));

  //update image
  glWidget->setQImage(image);
  QTimer::singleShot(10, glWidget, SLOT(update()));
}

void CamCalibPanel::on_matlabFileButton_clicked()
{
  //load
  if (!loadMatlabCalib())
  {
    QMessageBox::critical(this, "Error", 
      QString("Matlab calibration file '%1' load failed."
        " Check that the file exists in the current work directory.")
          .arg(MATLAB_CALIB_FILENAME));
  }
}

void CamCalibPanel::resetPanel(void)
{
  auto mainWin = getApp()->getMainWindow();
  if (mainWin) { mainWin->getGLWidget()->clearImage(); }
  _model.clear();
  _calib.clear();
  _reprojection.clear();
  calibGroupUpdate();
}

void CamCalibPanel::calibGroupUpdate()
{
  imageCountLine->setText(QString::number(_calib.imageCount));
  imageWidthLine->setText(QString::number(_calib.imageSize.width()));
  imageHeightLine->setText(QString::number(_calib.imageSize.height()));
  K00_Line->setText(QString::number(_calib.K(0, 0)));
  K01_Line->setText(QString::number(_calib.K(0, 1)));
  K02_Line->setText(QString::number(_calib.K(0, 2)));
  K10_Line->setText(QString::number(_calib.K(1, 0)));
  K11_Line->setText(QString::number(_calib.K(1, 1)));
  K12_Line->setText(QString::number(_calib.K(1, 2)));
  K20_Line->setText(QString::number(_calib.K(2, 0)));
  K21_Line->setText(QString::number(_calib.K(2, 1)));
  K22_Line->setText(QString::number(_calib.K(2, 2)));
  kc1Line->setText(QString::number(_calib.kc[0]));
  kc2Line->setText(QString::number(_calib.kc[1]));
  kc3Line->setText(QString::number(_calib.kc[2]));
  kc4Line->setText(QString::number(_calib.kc[3]));
  kc5Line->setText(QString::number(_calib.kc[4]));
}

bool CamCalibPanel::loadMatlabCalib(void)
{
  //reset GUI
  resetPanel();

  //calib file
  auto workDir = getApp()->getMainWindow()->getWorkDirectory();
  QDir dir(workDir);
  auto fileName = dir.absoluteFilePath(MATLAB_CALIB_FILENAME);

  //load
  bool rv = _calib.loadMatlabCalib(fileName);

  //update GUI
  calibGroupUpdate();  
  updateImageFiles();

  return rv;
}

void CamCalibPanel::updateMinMaxError(void)
{
  if (!_reprojection.size()) 
  { //no data
    minErrorLine->clear();
    maxErrorLine->clear();
    return;
  }

  double minErr = _reprojection.first().rmsError;
  double maxErr = _reprojection.first().rmsError;
  foreach (auto const& item, _reprojection)
  {
    if (item.rmsError<minErr) { minErr = item.rmsError; }
    if (item.rmsError>maxErr) { maxErr = item.rmsError; }
  }

  minErrorLine->setText(QString::number(minErr));
  maxErrorLine->setText(QString::number(maxErr));
}