// Software developed for the Spring 2016 Brown University course
// ENGN 2502 3D Photography
// Copyright (c) 2016, Daniel Moreno and Gabriel Taubin
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the Brown University nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL DANIEL MORENO NOR GABRIEL TAUBIN BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "CameraRecorder.hpp"

#include <QDir>
#include <QTime>
#include <QVariant>
#include <QApplication>

#include <cam/cam.hpp>
#include <img/util.hpp>
#include <img/yuv.hpp>

CameraRecorder::CameraRecorder(QObject * parent) :
  QThread(parent),
  _count(0),
  _interval(0),
  _path(),
  _namePattern("image_%03d.png"),
  _imageNames(),
  _rotation(0),
  _camera(),
  _cameraName(),
  _format(),
  _message(),
  _error(false),
  _skip(40),
  _index(0),
  _timer() {
  qDebug(" -- CameraRecorder created -- ");
}

CameraRecorder::~CameraRecorder() {
  qDebug(" -- CameraRecorder destroyed -- ");
}

void CameraRecorder::run() {
  _error = false;
  _imageNames.clear();
  _index = -_skip;

  emit started();

  if (_cameraName.isEmpty()) { //no camera
    stop(true, "Camera was not set");
    return;
  }

  //check if we have a camera
  if (_camera.isNull()) {
    stop(true, "Failed to get camera '" + _cameraName + "'");
    return;
  }

  //set camera format
  if (!_format.isEmpty()) {
    _camera->setCameraProperty(CAM_PROP_FORMAT, _format);
  }

  //connect
  connect(_camera.data(), &CameraInterface::newImage,
          this,           &CameraRecorder::on_camera_newImage, Qt::DirectConnection);

  //start the camera
  if (!_camera->start()) { //failed
    stop(true, "Failed to start camera '" + _cameraName + "'");
    return;
  }

  //start the image timer
  _timer.start();

  //start event loop
  exec();
}

void CameraRecorder::setCameraName(QString cameraName) { 
  _cameraName = cameraName; 
  if (!_cameraName.isEmpty()) {
    //get camera (do it here so that the camera is created in the caller's thread)
    _camera = cam::getCamera(_cameraName);
  }
}

void CameraRecorder::on_camera_newImage(ImageBuffer const& image) {
  if (_index<0) { //skip images
    ++_index;
    return;
  }

  if (_index >= _count) { //finished
    stop();
    return;
  }

  if (_interval && _timer.elapsed() < _interval) { //must wait, skip this image
    return;
  }

  qDebug("debug new image recorder");

  ImageBuffer flippedImage;
  if (image.format==ImageBuffer::NV12) {
    ImageBuffer rgbImage;
    if (img::rgbFromNV12(image, rgbImage)) {
      img::flipVertical(rgbImage, flippedImage);
    }
  } else { //I hope is RGB ...
    img::flipVertical(image, flippedImage);
  }
  
  ImageBuffer rotatedImage;
  if (_rotation!=0) {
    img::rotateImage(flippedImage, _rotation, rotatedImage);
  }

  auto & outImage = (rotatedImage.data?rotatedImage:flippedImage);

  //save the image
  QDir dir(_path);
  QString basename = QString::asprintf(qPrintable(_namePattern), _index);
  QString filename = dir.absoluteFilePath(basename);
  if (img::save(filename.toStdString(), outImage)) {
    _imageNames.append(basename);
    qDebug("saved image: %s", qPrintable(filename));
  } else {
    qDebug("saved image failed %d", _index);
    stop(true, QString("Save image %1 failed").arg(_index));
    return;
  }

  emit progress();
  _timer.restart();
  ++_index;
}

void CameraRecorder::stop(bool error, QString message) {

  if (!_camera.isNull()) { //stop
    disconnect(_camera.data(), &CameraInterface::newImage,
               this,           &CameraRecorder::on_camera_newImage);
    QThread::msleep(100);
    _camera.clear();
  }

  _error = error;
  _message = message;

  emit finished();
  quit();
}
