// Software developed for the Spring 2016 Brown University course
// ENGN 2502 3D Photography
// Copyright (c) 2016, Daniel Moreno and Gabriel Taubin
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the Brown University nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL DANIEL MORENO NOR GABRIEL TAUBIN BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "MainWindow.hpp"

#include <QSettings>
#include <QFileDialog>
#include <QMessageBox>

#include "Application.hpp"
#include "AboutDialog.hpp"

QMap<QString,std::function<QWidget*(QWidget *parent)>> & MainWindow::getRegisteredPanels(void)
{
  typedef QMap<QString,std::function<QWidget*(QWidget *parent)>> MapType;
  static MapType * _registeredPanels = new MapType;
  return *_registeredPanels;
}

bool MainWindow::registerPanel(QString panelName, std::function<QWidget*(QWidget *parent)> panelFactory)
{
   getRegisteredPanels().insert(panelName, panelFactory);
   return true;
}

MainWindow::MainWindow(QWidget * parent, Qt::WindowFlags flags) :
  QMainWindow(parent, flags)
{
  setupUi(this);

  setWindowTitle(APP_NAME);

  //fill the panel combo
  auto panelNames = getRegisteredPanels().keys();
  foreach (auto item, panelNames)
  {
    panelCombo->insertItem(999, item);
  }

  //restore settings
  QSettings config;
  restoreGeometry(config.value("MainWindow/geometry").toByteArray());
  restoreState(config.value("MainWindow/windowState").toByteArray());
  auto panel = config.value("MainWindow/panel").toString().trimmed();
  if (!panel.isEmpty() && panelNames.contains(panel)) { panelCombo->setCurrentText(panel); }
  workDirLine->setText(config.value("MainWindow/workDirectory", QDir::currentPath()).toString());
  workDirUpdated();
}

void MainWindow::closeEvent(QCloseEvent *e)
{
  //save settings
  QSettings config;
  config.setValue("MainWindow/geometry", saveGeometry());
  config.setValue("MainWindow/windowState", saveState());
  config.setValue("MainWindow/panel", panelCombo->currentText());
  config.setValue("MainWindow/workDirectory", workDirLine->text());

  QMainWindow::closeEvent(e);
}

void MainWindow::on_actionAbout_triggered()
{
  AboutDialog d;
  d.exec();
}

void MainWindow::on_panelCombo_currentIndexChanged(int index)
{
  qDebug("MainWindow::on_panelCombo_currentIndexChanged: index %d", index);
  setPanel(panelCombo->currentText());
}

void MainWindow::setPanel(QString name)
{
  //remove current widgets from panel
  auto children = currPanelWidget->children();
  foreach(auto child, children)
  {
    auto widget = dynamic_cast<QWidget*>(child);
    if (widget)
    {
      widget->hide();
      widget->deleteLater();
    }
  }

  //delete current layout
  auto oldLayout = currPanelWidget->layout();
  if (oldLayout)
  {
    delete oldLayout;
  }

  //create a new layout
  QVBoxLayout * newLayout = new QVBoxLayout(currPanelWidget);
  newLayout->setContentsMargins(0, 0, 0, 0);

  //add new panel
  QWidget * panel = nullptr;
  
  auto & registeredPanels = getRegisteredPanels();
  auto panelFactory = registeredPanels.find(name);
  if (panelFactory!= registeredPanels.end())
  { //panel found
    panel = (*panelFactory)(currPanelWidget);
  }

  if (!panel)
  { //panel not found
    glwidget->clearImage();
    qDebug("MainWindow::on_panelCombo_currentIndexChanged: panel '%s' not found", qPrintable(name));
    return;
  }

  //setup new panel
  newLayout->addWidget(panel);
  panel->show();
}

void MainWindow::on_workDirButton_clicked()
{
  QString dir = QFileDialog::getExistingDirectory(this, "Select Work Directory", workDirLine->text(), QFileDialog::ShowDirsOnly | QFileDialog::DontResolveSymlinks);
  if (!dir.isEmpty())
  {
    workDirLine->setText(dir);
    workDirUpdated();
  }
}

void MainWindow::on_workDirLine_editingFinished()
{
  workDirUpdated();
}

void MainWindow::workDirUpdated(void)
{
  static QMutex mutex;
  if (!mutex.tryLock(0))
  {
    return;
  }

  QString dirName = workDirLine->text();;
  QString prevDirName = workDirLine->property("prevValue").toString();

  if (prevDirName.isEmpty())
  {
    prevDirName = QDir::currentPath();
  }
  if (dirName.isEmpty())
  {
    dirName = prevDirName;
  }

  QDir workDir(dirName);
  if (!workDir.exists())
  { //doesn't exist: ask to create
    auto button = QMessageBox::question(nullptr, "Warning", "Directory '" + dirName + "' does not exist.\nDo you want to create it?");
    if (button == QMessageBox::Yes)
    {
      if (workDir.mkpath(dirName))
      {
        qInfo("Directory created: %s", qPrintable(dirName));
      }
      else
      { //couldn't create
        qCritical("Directory create failed: %s", qPrintable(dirName));
        dirName = prevDirName;
      }
    }
    else
    { //user cliked 'No'
      dirName = prevDirName;
    }
  }//workDir exists

  //update the value
  if (workDirLine->text() != dirName)
  {
    workDirLine->setText(dirName);
  }
  workDirLine->setProperty("prevValue", dirName);
  emit workDirectoryChanged();
  mutex.unlock();
}

QString MainWindow::getWorkDirectory(void) const
{
  return workDirLine->text();
}

void MainWindow::setChangesEnabled(bool enabled)
{
  panelCombo->setEnabled(enabled);
  workDirLabel->setEnabled(enabled);
  workDirLine->setEnabled(enabled);
  workDirButton->setEnabled(enabled);
}